// #name = RoboRoute (Server)
// #version = 0.3
// #date = 22 September 2002
// #author = {BEE}Shoddy[KE]
// #warrior = Shoddy
// #email = erdlftz@yahoo.com
// #web = http://www.beetribe.com/
// #author = SquirrelOfDeath
// #warrior = SquirrelOfDeath 
// #email = squirrel@teambro.org
// #web = http://sod.tribes2.org/
// #description = Record routes and have bots play them back.
// #status = Beta

// =======================================================

$RoboRouteDir = "RoboRoute/";
$RoboRoute::Skill = 0.75;

// These period values must divide 1000 evenly for position snapshots to 
// be taken
$RoboRoute::RecPeriod = 50; // delay between snapshots when on foot
$RoboRoute::VehicleRecPeriod = 50; // delay between snapshots when in vehicle

// Maximum distance from desired position before bot's position is forced
$RoboRoute::MaxPosError = 10; 

// Bot created before start of match wait between 0 and this many seconds
// before joining game (to avoid all joining at the same time).
$RoboRoute::MatchStartDelay = 15; 

// Do not change this value
$RoboRoute::LegacyRecPeriod = 50;

// Should bots die if they get bodyblocked?
$RoboRoute::LethalBB = true;

// Play sound when player bodyblocks a bot?
$RoboRoute::BBSound = true;

// Speed at which a bot has to be moving to be able to be bodyblocked
$RoboRoute::MinBBSpeed = 50/3.6; // 50 kph

// State vars, do not touch
if ( $RoboRoute::Recording $= "" ) $RoboRoute::Recording = false;
if ( $RoboRoute::AIEnabled $= "" ) $RoboRoute::AIEnabled = true;
if ( !isObject( $RoboRoute::BotGroup ) ) {
    $RoboRoute::BotGroup = new SimGroup();
}

function min( %a, %b ) { return ( %a <= %b ? %a : %b ); }
function max( %a, %b ) { return ( %a >= %b ? %a : %b ); }

function BotInfo::spawn( %this ) {
    %routedir = $RoboRouteDir @ $MissionName @ "/" @ %this.team @ "/" @ 
	 %this.group @ "/";

    if ( %this.route $= "" ) {
	%routeList = RoboRouteGetRouteList( %this.team, %this.group );
	%numRoutes = getFieldCount( %routeList );
	%routeNum = mFloor( getRandom() * %numRoutes );
	%route = getField( %routeList, %routeNum );
    } else {
	%route = %this.route;
    }

    deleteVariables( "$BotInfo::*" );
    $BotInfo::Period = 0;
    exec( %routedir @ %route @ ".cs" );

    if ( $BotInfo::nVels $= "" ) {
	%this.routeData.delete();
	%this.routeData = __RoboRouteData__.getID();
	if ( !isObject( %this.routeData ) ) {
	    messageClient(%client, "MsgClient", "\c2Error loading route. Removing bot \c5" @ %this.name @ "\c2.");

	    %this.remove();
	    return;
	}
    
	%this.routeData.setName("");
	%routeData = %this.routeData;
    } else {
	// For backwards compatibility
	if ( $BotInfo::Period == 0 ) {
	    $BotInfo::Period = $RoboRoute::LegacyRecPeriod;
	}

	%routeData = %this.routeData;

	%routeData.InVehicle = $BotInfo::InVehicle;
	%routeData.Vehicle = $BotInfo::Vehicle;
	%routeData.StartTrans = $BotInfo::StartLoc;
	%routeData.StartVel = $BotInfo::StartVel;
	%routeData.nVels = $BotInfo::nVels;
	%routeData.ArmorSize = $BotInfo::ArmorSize;
	%routeData.Period = $BotInfo::Period;
	
	if ( %routeData.InVehicle ) {
	    for (%i=0; %i<%routeData.nVels; %i++) {
		%routeData.Trans[%i] = $BotInfo::Trans[%i];
	    }
	} else {
	    for (%i=0; %i<%routeData.nVels; %i++) {
		%routeData.Vel[%i] = $BotInfo::Vel[%i];
	    }
	}
    }

    if ( %routeData.Period < 10 ) {
	// Enforce minimum period to prevent from freezing server
	%routeData.Period = 10;
    }

    %this.client = AIConnect(%this.name, %this.team, $RoboRoute::Skill, true);
    %this.client.isRoboRouteBot = true;
    %BotPlayer = %this.client.player;
    %BotPlayer.SetArmor(%routeData.ArmorSize);

    // Set inventory if route has inventory info
    if ( %routeData.weapon[0] !$= "" ) {
	%routeData.setInventory( %BotPlayer );
    }
 
    if ( %routeData.InVehicle ) {
        %routeData.vehicle = RoboRouteSpawnVehicle( %routeData.Vehicle, 
	    %BotPlayer, %this.team, %routeData.StartTrans );
        %this.schedule(%routeData.Period, "MoveVehicle");
    } else {
	%BotPlayer.setTransform( %routeData.StartTrans );
	%BotPlayer.setVelocity( %routeData.StartVel );
	%this.client.pressJump();
	%this.client.pressJet();

	%this.schedule(%routeData.Period, "MoveBot");
    }

    %routeData.curIdx = 0;
    %routeData.velCorrection = "0 0 0";
}

function BotInfo::MoveVehicle( %this ) {
    %routeData = %this.routeData;
    if (isObject(%routeData.vehicle)) {
	%vehicle = %routeData.vehicle;
        %routeData.curIdx += 1;
	if (%routeData.curIdx <= %routeData.nVels) {
            %vehicle.setTransform(%routeData.Trans[%routeData.curIdx]);
            %this.schedule(%routeData.Period, "MoveVehicle");
	    return;
	}
        else {
	    %this.client.player.scriptKill(0);
	}
    }
    %this.client.drop();
    %this.client = 0;
    %this.FinishedRoute();
}

function BotInfo::MoveBot( %this ) {
    %routeData = %this.routeData;
    if (isObject(%this.client.player)) {
	%player = %this.client.player;
        %routeData.curIdx += 1;
	if (%routeData.curIdx <= %routeData.nVels) {
	    %this.client.pressJump();
	    %this.client.pressJet();

	    if ( %routeData.Trans[%routeData.curIdx] !$= "" ) {
		%this.CorrectPosition( %routeData.Trans[%routeData.curIdx] );
	    }

	    %vel = VectorAdd( %routeData.Vel[%routeData.curIdx],
			      %this.routeData.velCorrection );

            %player.setVelocity(%vel);

            %this.schedule(%routeData.Period, "MoveBot");
	    return;
	}
        else {
	    %player.scriptKill(0);
	}
    }
    %this.client.drop();
    %this.client = 0;
    %this.FinishedRoute();
}

function BotInfo::CorrectPosition( %this, %trans ) {
    %mypos = %this.client.player.getPosition();
    %recpos = getWords( %trans, 0, 2 );
    %vec = VectorSub( %recpos, %mypos );    
    %dist = VectorDist( %vec, "0 0 0" );
    if ( %dist >= $RoboRoute::MaxPosError ) {
	%this.client.player.setTransform(%trans);
	%this.routeData.velCorrection = "0 0 0";
    } else {
	%this.routeData.velCorrection = %vec;
    }
}

function BotInfo::FinishedRoute( %this ) {
    %this.loopsDone += 1;
    %this.routeData.delete();

    if ( %this.loopCount == 0 || %this.loopsDone < %this.loopCount ) {
	%this.routeData = new ScriptObject() {
	    class = RouteInfo;
	};
	%this.respawnTimer = %this.schedule( %this.loopDelay * 1000, "Spawn" );
    } else {
	messageAll(%client, "\c2Route sequence completed for " @ %this.name  
	    @ ".");
	$RoboRoute::BotGroup.remove( %this );
	%this.delete();
    }
}

function BotInfo::remove( %this ) {
    if ( isObject( %this.client ) ) {
	if ( isObject( %this.client.player ) ) {
	    %this.client.player.scriptKill(0);
	}
	%this.client.drop();
	%this.client = 0;
    } else {
	if ( isObject(%this.respawnTimer) ) {
	    %this.respawnTimer.cancel();
	}
    }
    if ( isObject( %this.routeData ) ) {
	%this.routeData.delete();
    }

    $RoboRoute::BotGroup.remove( %this );

    %this.delete();
}

function ServerCmdRoboRouteGetBotList( %client ) {
    %count = $RoboRoute::BotGroup.getCount();
    %BotList = "";
    %GroupList = "";
    %RouteList = "";
    for (%i=0; %i<%count; %i++) {
	%o = $RoboRoute::BotGroup.getObject(%i);
	if ( %i>0 ) {
	    %BotList = %BotList @ "\t";
	    %GroupList = %GroupList @ "\t";
	    %RouteList = %RouteList @ "\t";
	}
	%BotList = %BotList @ %o.name;
	%GroupList = %GroupList @ %o.group;
	%RouteList = %RouteList @ %o.route;
    }

    commandToClient( %client, 'RoboRouteSetBotList', %BotList, %GroupList, %RouteList );
}

function ServerCmdRoboRouteRemoveBot( %client, %name ) {
    if ( !%client.isAdmin ) {
	messageClient(%client, 'MsgClient', "\c2You must be an admin to remove a bot.");
	return;
    }

    %botinfo = RoboRouteGetBotInfo( %name );
    if ( isObject(%botinfo) ) { 
	%botinfo.remove();
	messageAll('MsgClient', "\c2" @ detag(getTaggedString(%client.name)) @ 
		   " removed bot \c5" @ %name @ "\c2.");
    } else {
	messageClient(%client, 'MsgClient', "\c2Bot \c5" @ %name @ "\c2 does not exist.");
    }
}

function ServerCmdRoboRoutePull( %client, %team, %group, %route, %BotName,
				 %loopCount, %loopDelay ) {
    if ( !%client.isAdmin ) {
	messageClient(%client, "MsgClient", "\c2You must be an admin to create a bot.");
	return;
    }

    if ( %team < 1 || %team > 2 ) {
	messageClient(%client, "MsgClient", "\c2Invalid team.");
	return;
    }

    // Make sure route exists, or if random route, that group has routes
    %routedir = $RoboRouteDir @ $MissionName @ "/" @ %team @ "/" @ %group @ "/";
    if ( %route $= "" ) {
	// Random route
	%routeList = RoboRouteGetRouteList( %team, %group );
	if ( getFieldCount( %routeList ) == 0 ) {
	    messageClient(%client, "MsgClient", "\c2Group has no routes.");
	    return;
	}
    } else {
	%filename = %routedir @ %route @ ".cs";
	if ( !isFile( %filename ) ) {
	    messageClient(%client, "MsgClient", "\c2File " @ %filename @ " not found.");
	    return;
	}
    }

    if ( %loopCount < 0 ) {
	messageClient(%client, "MsgClient", "\c2Invalid loop count.");
	return;
    }

    if ( %loopDelay < 0 ) {
	messageClient(%client, "MsgClient", "\c2Invalid loop delay.");
	return;
    }

    if ( %BotName $= "" ) {
	%BotName = "Bot";
    }

    %BotNameRoot = %BotName;
    %idx = 1;
    while ( RoboRouteGetBotInfo( %BotName ) != 0 ) {
	%BotName = %BotNameRoot @ %idx;
	%idx += 1;
    }

    %botObj = new ScriptObject() {
	class = BotInfo;
	name = %BotName;
	team = %team;
	group = %group;
	route = %route;
	loopCount = %loopCount;
	loopDelay = %loopDelay;
	client = 0;
	loopsDone = 0;
    };
    %botObj.routeData = new ScriptObject() {
	class = RouteInfo;
    };

    $RoboRoute::BotGroup.add( %botObj );
    
    messageAll('MsgClient', "\c2" @ detag(getTaggedString(%client.name)) @ 
	       " created bot \c5" @ %botObj.name @ "\c2.");

    if ( $MatchStarted ) {
	%botObj.spawn();
    }
}

function RoboRouteGetBotInfo( %name ) {
    %count = $RoboRoute::BotGroup.getCount();
    for (%i=0; %i<%count; %i++) {
	%o = $RoboRoute::BotGroup.getObject(%i);
	if ( %o.name $= %name ) {
	    return %o;
	}
    }
    return 0;
}

function RoboRouteSpawnVehicle( %vehicle, %pilot, %team, %transform ) {
    switch$( %vehicle ) {
    case "Shrike": %datablock = "ScoutFlyer";
    case "Bomber": %datablock = "BomberFlyer";
    case "Tank":   %datablock = "AssaultVehicle";
    case "Bike":   %datablock = "ScoutVehicle";
    case "MPB":    %datablock = "MobileBaseVehicle";
    case "HAPC":   %datablock = "HAPCFlyer";
    default:       %datablock = %vehicle;
    }

    %obj = %datablock.create( %team );
    if ( %obj ) {
        %obj.team = %team;
        vehicleListAdd(%blockName, %obj);
	MissionCleanup.add(%obj);

        %obj.setTransform( %transform );
        
        %obj.lastPilot = %pilot;
        %pilot.lastVehicle = %obj;
        %pilot.client.usingVehicle = true;
        %pilot.client.pilotVehicle = true;

        %obj.getDatablock().schedule( 200, "mountDriver", %obj, %pilot );

        if(%obj.getTarget() != -1) {
	    setTargetSensorGroup(%obj.getTarget(), %team);
        }
    }

    return %obj;
}

function ServerCmdRoboRouteStartRecord(%client, %group, %desc) {
    if ( !RoboRouteIsValidFilename( %group ) ||
	 !RoboRouteIsValidFilename( %desc ) ) 
    {
	messageClient(%client, "MsgClient", "\c2Invalid filename.");
	return;
    }

    if ($RoboRoute::Recording) {
	messageClient(%client, "MsgClient", "\c2A recording is already running.");
	return;
    }

    if (!%client.isSuperAdmin) {
	messageClient(%client, "MsgClient", "\c2You must be a super admin to record.");
	return;
    }

    if (!isObject(%client))
	return;

    %player = %client.player;
    if (!isObject(%player))
	return;

    %routeData = new ScriptObject() {
	class = RouteInfo;
    };

    %routeData.InVehicle = %player.isMounted();
    if ( %routeData.InVehicle ) {
	%routeData.Vehicle = %player.lastVehicle.getDatablock().getName();
    }
    %routeData.StartTrans = %player.getTransform();
    %routeData.StartVel = %player.getVelocity();
    %routeData.nVels = 0;
    %routeData.ArmorSize = %player.getArmorSize();
    %routeData.saveInventory( %client.player );
    %routeData.Period = ( %routeData.InVehicle ? $RoboRoute::VehicleRecPeriod : $RoboRoute::RecPeriod );
    schedule( %routeData.Period, 0, "RoboRouteContinueRecord", %client);
    $RoboRoute::Recording = true;
    $RoboRoute::RecGroup = %group;
    $RoboRoute::RecRouteDesc = %desc;
    $RoboRoute::RecTeam = %client.getSensorGroup();
    $RoboRoute::RecData = %routeData;
    $RoboRoute::RecClient = %client;

    messageClient(%client, "MsgClient", "\c2Bot recording started.");
}

function RouteInfo::saveInventory( %this, %player ) {
    // Weapons
    for (%i=0; $InvWeapon[%i] !$= ""; %i++) {
	%inv = $NameToInv[$InvWeapon[%i]]; 
	%count = %player.getInventory( %inv );
	if ( %count $= "" ) %count = 0;
	%this.weapon[%i] = %count;
    }

    // Packs
    for (%i=0; $InvPack[%i] !$= ""; %i++) {
	%inv = $NameToInv[$InvPack[%i]]; 
	%count = %player.getInventory( %inv );
	if ( %count $= "" ) %count = 0;
	%this.pack[%i] = %count;
    }

    // Grenades
    for (%i=0; $InvGrenade[%i] !$= ""; %i++) {
	%inv = $NameToInv[$InvGrenade[%i]]; 
	%count = %player.getInventory( %inv );
	if ( %count $= "" ) %count = 0;
	%this.grenade[%i] = %count;
    }

    // Mines
    for (%i=0; $InvMine[%i] !$= ""; %i++) {
	%inv = $NameToInv[$InvMine[%i]]; 
	%count = %player.getInventory( %inv );
	if ( %count $= "" ) %count = 0;
	%this.mine[%i] = %count;
    }
}

function RouteInfo::setInventory( %this, %player ) {

    %player.clearInventory();
    %weaponcount = 0;

    // Weapons
    for (%i=0; %this.weapon[%i] !$= ""; %i++) {
	%inv = $NameToInv[$InvWeapon[%i]]; 
	%player.setInventory( %inv, %this.weapon[%i] );
	if ( %this.weapon[%i] > 0 ) {
	    %weaponcount += 1;
	}

	switch$ ( %inv ) {
	case Plasma:
	    %player.setInventory( PlasmaAmmo, 400 );  
	case Disc:
	    %player.setInventory( DiscAmmo, 400 );
	case GrenadeLauncher:
	    %player.setInventory( GrenadeLauncherAmmo, 400 );
	case Mortar:
	    %player.setInventory( MortarAmmo, 400 );
	case MissileLauncher:
	    %player.setInventory( MissileLauncherAmmo, 400 );
	case Chaingun:
	    %player.setInventory( ChaingunAmmo, 400 );
	}
    }

    // Packs
    for (%i=0; %this.pack[%i] !$= ""; %i++) {
	%inv = $NameToInv[$InvPack[%i]]; 
	%player.setInventory( %inv, %this.pack[%i] );
	if ( %this.pack[%i] > 0 ) {
	    // Can only have 1 pack
	    %pack = %inv;
	    break;
	}
    }

    // Grenades
    for (%i=0; %this.grenade[%i] !$= ""; %i++) {
	%inv = $NameToInv[$InvGrenade[%i]]; 
	%player.setInventory( %inv, %this.grenade[%i] );
	%player.lastGrenade = %inv;
    }

    // Mines
    for (%i=0; %this.mine[%i] !$= ""; %i++) {
	%inv = $NameToInv[$InvMine[%i]]; 
	%player.setInventory( %inv, %this.mine[%i] );
    }

    %player.setInventory( RepairKit, 1 );
    %player.setInventory( Beacon, 400 );
    %player.setInventory( TargetingLaser, 1 );

    // hack from buyFavorites()
    if ( %pack $= "AmmoPack" ) {
	invAmmoPackPass(%player.client);
    }
}

function RoboRouteContinueRecord(%client) {
    if ($RoboRoute::Recording) {
	if (isObject(%client)) {
            %player = %client.player;
            if ( $RoboRoute::RecData.InVehicle ) {
                if ( isObject( %player ) && isObject( %player.lastVehicle ) ) {
                    $RoboRoute::RecData.nVels += 1;
		    $RoboRoute::RecData.Trans[$RoboRoute::RecData.nVels] = %player.lastVehicle.getTransform();

		    schedule( $RoboRoute::RecData.Period, 0, "RoboRouteContinueRecord", %client);
                    return;
                }
            } else {
		if (isObject(%player)) {
		    $RoboRoute::RecData.nVels += 1;
                    $RoboRoute::RecData.Vel[$RoboRoute::RecData.nVels] = %player.getVelocity();
		    // Record Position & Orientation every second
		    if ( $RoboRoute::RecData.nVels % 
			 (1000 / $RoboRoute::RecData.Period) == 0 ) 
		    {
			$RoboRoute::RecData.Trans[$RoboRoute::RecData.nVels] =
			    %player.getTransform();
		    }

                    schedule( $RoboRoute::RecData.Period, 0, "RoboRouteContinueRecord", %client);
		    return;
		}
            }
	}
	$RoboRoute::Recording = false;
	if ( isObject( $RoboRoute::RecData ) ) {
	    $RoboRoute::RecData.delete();
	}
	$RoboRoute::RecData = 0;
    }
}

function ServerCmdRoboRouteEndRecord(%client) {
    if (!$RoboRoute::Recording) {
	messageClient(%client, "MsgClient", "\c2No recording in progress.");
	return;
    }

    if (%client != $RoboRoute::RecClient) {
	messageClient(%client, "MsgClient", "\c2You cannot stop another client's recording.");
	return;
    }

    %team = %client.getSensorGroup();

    $RoboRoute::Recording = false;
    $RoboRoute::RecClient = 0;

    %fileNameRoot = $RoboRouteDir @ $MissionName @ "/" @ $RoboRoute::RecTeam @
	 "/" @ $RoboRoute::RecGroup @ "/" @ $RoboRoute::RecRouteDesc;
    %fileName = %fileNameRoot @ ".cs";

    $RoboRoute::RecData.setName("__RoboRouteData__");
    $RoboRoute::RecData.save( %filename );

    messageClient(%client, "MsgClient", "\c2Bot recording saved to " @ %fileName @ ".");

    %fileName = %fileNameRoot @ ".dso";
    deleteFile(%fileName);

    // Clear the data
    $RoboRoute::RecData.delete();
    $RoboRoute::RecData = 0;
}

function ServerCmdRoboRouteGetGroupList(%client, %team) {
    if ( %team < 1 || %team > 2 ) {
	messageClient(%client, 'MsgClient', "\c2Invalid team.");
	return;
    }

    %groupList = RoboRouteGetGroupList(%team);

    commandToClient( %client, 'RoboRouteSetGroupList', %groupList );
}

function RoboRouteGetGroupList(%team) {
    if ( $MissionName $= "" ) {
	error( "Not in a mission" );
	return;
    }

    %pattern = $RoboRouteDir @ $MissionName @ "/" @ %team @ "/*";
    %patternlen = strlen( %pattern ) - 1;

    %groupList = "";
    %numGroups = 0;
    for ( %file = findFirstFile( %pattern );
	  %file !$= "";
	  %file = findNextFile( %pattern ) ) 
    {
	%group = getSubStr( %file, %patternlen, strlen(%file) );
	%slashpos = strpos( %group, "/" );
	if ( %slashpos <= 0 ) {
	    // No group found
	    continue;
	}
	%group = getSubStr( %group, 0, %slashpos );

	for (%i=0; %i<%numGroups; %i++) {
	    if ( %group $= getField( %groupList, %i ) ) {
		break;
	    }
	}

	if ( %i < %numGroups ) {
	    // Not a new group
	    continue;
	}

	if ( %groupList !$= "" ) {
	    %groupList = %groupList @ "\t";
	}
	%groupList = %groupList @ %group;

	%lastGroup = %group;
	%numGroups += 1;
    }

    return %groupList;
}

function RoboRouteMakeNewGroup(%team, %group) {
    if ( $MissionName $= "" ) {
	error( "Not in a mission" );
	return false;
    }

    %file = new FileObject();

    %success = %file.OpenForWrite( $RoboRouteDir @ $MissionName @ "/" @ %team @ "/" @ %group @ "/dummy.txt" );

    if ( !%success ) {
	%file.delete();
	return false;
    }

    %file.WriteLine( "This file is needed for the proper operation of the RoboRoute script. Please do not delete." );

    %file.close();
    %file.delete();

    return true;
}

function ServerCmdRoboRouteMakeNewGroup(%client, %team, %group) {
    if ( %group $= "" || !RoboRouteIsValidFilename(%group) ) {
	messageClient(%client, 'MsgClient', "\c2Invalid group name.");
	return;
    }

    if ( %team < 0 || %team > 2 ) {
	messageClient(%client, 'MsgClient', "\c2Invalid team.");
	return;
    }

    %success = RoboRouteMakeNewGroup( %team, %group );

    if ( %success ) {
	messageClient(%client, 'MsgClient', "\c2Group '" @ %group @ "' created.");
	return;
    } else {
	messageClient(%client, 'MsgClient', "\c2Error creating group '" @ %group @ "'.");
	return;
    }
}

function ServerCmdRoboRouteGetRouteList( %client, %team, %group ) {
    if ( %team < 1 || %team > 2 ) {
	messageClient(%client, 'MsgClient', "\c2Invalid team.");
	return;
    }

    %routeList = RoboRouteGetRouteList( %team, %group );

    commandToClient( %client, 'RoboRouteSetRouteList', %routeList );
}

function RoboRouteGetRouteList( %team, %group ) {
    %pattern = $RoboRouteDir @ $MissionName @ "/" @ %team @ "/" @ %group
	 @ "/*.cs";
    %patternlen = strlen(%pattern) - 4;

    %routeList = "";

    for ( %file = findFirstFile( %pattern );
	  %file !$= "";
	  %file = findNextFile( %pattern ) ) 
    {
	%route = getSubStr( %file, %patternlen, strlen(%file) );

	// Remove .cs
	%route = getSubStr( %route, 0, strlen(%route)-3 );

	if ( %routeList !$= "" ) {
	    %routeList = %routeList @ "\t";
	}
	%routeList = %routeList @ %route;
    }

    return %routeList;
}

function RoboRouteSetAI( %ai_state ) {
    $RoboRoute::AIEnabled = ( %ai_state ? true : false );
    AISystemEnabled( $RoboRoute::AIEnabled );

}
    

// Left in for backwards-compatibility with old clients.  New clients use
// ServerCmdRoboRouteSetOption.
function ServerCmdRoboRouteToggleAI( %client ) {
    if ( !%client.isAdmin ) {
	messageClient(%client, 'MsgClient', "\c2You must be an admin to toggle the AI.");
	return;
    }

    RoboRouteSetAI( !$RoboRoute::AIEnabled );
}

function ServerCmdRoboRouteSetOption( %client, %option, %value ) {
    if ( !%client.isAdmin ) {
	messageClient(%client, 'MsgClient', "\c2You must be an admin to set this option.");
	return;
    }

    switch$ ( %option ) {
    case "AIEnabled":
	RoboRouteSetAI( %value );

    case "AISkill":
	$RoboRoute::Skill = min( 1.0, max( 0.0, %value ) );

    case "LethalBB":
	$RoboRoute::LethalBB = ( %value ? true : false );

    case "BBSound":
	$RoboRoute::BBSound = ( %value ? true : false );

    case "MinBBSpeed":
	$RoboRoute::MinBBSpeed = max( 0.0, %value/3.6 );

    default:
	messageClient(%client, 'MsgClient', "\c2Error: unrecognized option '" 
		      @ %option @ "'.");
	return;
    }

    messageAll('MsgClient', "\c2" @ detag(getTaggedString(%client.name)) @ 
	       " changed the value of RoboRoute option '" @ %option @ 
	       "' to " @ %value @ ".~wfx/misc/warning_beep.wav");
}

function ServerCmdRoboRouteGetOptions( %client ) {
    commandToClient( %client, 'RoboRouteSetOptions', 
		     $RoboRoute::AIEnabled, 
		     $RoboRoute::Skill, 
		     $RoboRoute::LethalBB, 
		     $RoboRoute::BBSound,
		     $RoboRoute::MinBBSpeed * 3.6 );
}

package RoboRouteServer {
    function DefaultGame::gameOver( %game ) {
	parent::gameOver( %game );

	while ( $RoboRoute::BotGroup.getCount() > 0 ) {
	    %o = $RoboRoute::BotGroup.getObject(0);
	    if ( isObject( %o ) && %o.class !$= "BotInfo" ) {
		%o.remove();
	    } 
	    $RoboRoute::BotGroup.remove( %o );
	}
    }

    function DefaultGame::startMatch( %game ) {
	parent::startMatch( %game );

	%count = $RoboRoute::BotGroup.getCount();
	for (%i=0; %i<%count; %i++) {
	    %o = $RoboRoute::BotGroup.getObject(%i);
	    %o.schedule( getRandom() * $RoboRoute::MatchStartDelay * 1000, 
			 "spawn" );
	}

    }

    function Armor::onCollision(%this,%obj,%col,%forceVehicleNode) {

	parent::onCollision( %this, %obj, %col, %forceVehicleNode );

	if ( !$RoboRoute::LethalBB && !$RoboRoute::BBSound ) {
	    // No body-blocking options selected, no poitn in continuing.
	    return;
	}

	if ( %obj.getState() $= "Dead" ) {
	    return;
	}

	%client = %obj.client;

	if ( !%obj.client.isRoboRouteBot ) {
	    return;
	} 

	%dataBlock = %col.getDatablock();
	%className = %dataBlock.className;

	if ( %className !$= "Armor" || %col.getState() $= "Dead" ||
	     %col.team == %obj.team || %col.isRoboRouteBot ) 
	{
	    return;
	}

	%vel1 = %client.player.getVelocity();
	%vel2 = %col.client.player.getVelocity();

	// bot has to be moving at least $RoboRoute::MinBBSpeed m/s
	if ( vectorLen( %vel1 ) < $RoboRoute::MinBBSpeed ) {
	    return;
	}

	// Get p2's speed relative to p1
	%rvel = vectorSub( %vel2, %vel1 );

	// Get unit-length direction of travel for p1
	%dir = vectorNormalize( %vel1 );

	// Project rvel onto p1's direction of travel
	%impactSpeed = -vectorDot( %rvel, %dir );

	// Speed of impact must be at least 1.5 m/s (arbitrary, but seems
	// to work pretty well)
	if ( %impactSpeed < 1.5 ) {
	    return;
	}

	if ( $RoboRoute::LethalBB ) {
	    Armor::damageObject( %this, %obj, %col, 
	                         %client.player.getPosition(),
			         99999, $DamageType::Ground, "0 0 0", 0 );
	}

	if ( $RoboRoute::BBSound ) {
	    messageClient( %col.client, 'BodyBlock', 
	                   "~wfx/weapons/plasma_rifle_projectile_hit.wav" );
	}
    }
};

activatePackage(RoboRouteServer);
